
// modified tga player code for SameGame

/*
 * tga player: plays an 8bpp .TGA file back into a buffer

 * Copyright (c) 1996-2002 VM Labs, Inc.  All rights reserved.
 *
 * NOTICE: VM Labs permits you to use, modify, and distribute this file
 * in accordance with the terms of the VM Labs license agreement
 * accompanying it. If you have received this file from a source other
 * than VM Labs, then your use, modification, or distribution of it
 * requires the prior written permission of VM Labs.
*/


// make sure there is no DEBUG since we do not want to call any printf
// statements
#undef DEBUG

#define DEBUG 1

#ifdef DEBUG
#include <stdio.h>
#endif
#include <stdlib.h>
#include <string.h>
#include <nuon/video.h>
#include <nuon/dma.h>
#include <nuon/bios.h>

#include "tga.h"

#ifndef TRUE
#define TRUE 1
#define FALSE 0
#endif

#define kBlackff 0x108080ff

static char *localScratch = NULL;  // Local RAM
static int localSize;

//****************************************************************
// some utility functions
//----------------------------------------------------------------

/* min and max macros */
#ifndef min
#define min(x,y)        ((x) < (y) ? (x) : (y))
#define max(x,y)        ((x) > (y) ? (x) : (y))
#endif

/* Bound an expression to the unit interval [low, high] */
#define Bound(low, high, x)      max((low), min((high), (x)))

int tColor = 0xff00ff00;    /* RGB !!! */

static unsigned long
My_YCCAFromRGBA(unsigned char r, unsigned char g, unsigned char b, unsigned char a)
{
  double rf, gf, bf;
  double yf, crf, cbf;
  int yk, crk, cbk;
  unsigned long color;

  if (((int)r == ((tColor>>24)&0xff)) && ((int)g == ((tColor>>16)&0xff)) && ((int)b == ((tColor>>8)&0xff))){
    color = 0x10808000;
  }else{

    // convert to YCCA format

    rf = (double)r/255.0;
    gf = (double)g/255.0;
    bf = (double)b/255.0;
    
    yf = Bound(0.0, 1.0, 0.299 * rf + 0.587 * gf + 0.114 * bf);
    crf = Bound(-0.5, 0.5, 0.500 * rf - 0.419 * gf - 0.081 * bf);
    cbf = Bound(-0.5, 0.5, -0.169 * rf - 0.331 * gf + 0.500 * bf);

    /* 0.5 added below to force rounding */
    yk = 16 + (int)(yf * 219.0 + 0.5);
    crk = 128 + (int)(crf * 224.0 + 0.5);
    cbk = 128 + (int)(cbf * 224.0 + 0.5);

//    color = (yk<<24) | (crk << 16 ) | (cbk << 8) | (255-a);
    color = (yk<<24) | (crk << 16 ) | (cbk << 8) | (a);
  }
    return color;
}


static unsigned long
YCC16FromRGB(unsigned char r, unsigned char g, unsigned char b)
{
  double rf, gf, bf;
  int yk, crk, cbk;
  unsigned long color;
  double yf, crf, cbf;

  rf = (double)r/255.0;
  gf = (double)g/255.0;
  bf = (double)b/255.0;

  yf = Bound(0.0, 1.0, 0.299 * rf + 0.587 * gf + 0.114 * bf);
  crf = Bound(-0.5, 0.5, 0.500 * rf - 0.419 * gf - 0.081 * bf);
  cbf = Bound(-0.5, 0.5, -0.169 * rf - 0.331 * gf + 0.500 * bf);

  /* 0.5 added below to force rounding */
  yk = (16 + (int)(yf * 219.0 + 0.5)) >> 2;
  crk = (128 + (int)(crf * 224.0 + 0.5)) >> 3;
  cbk = (128 + (int)(cbf * 224.0 + 0.5)) >> 3;

  color = (yk<<26) | (crk << 21 ) | (cbk << 16);
  return color;
}


#define PALETTE_SIZE 256
#define MAX_WIDTH 720

// tga points to the first byte of the source tga file
// dmaflags is the dmaflags of destination image
// base is the address of destination image
// xOutOffset is the x position of source in destination
// yOutOffset is the y position of source in destination
// clut is pointer to 256 longs where the clut of the source
// tga file will be put (this can be a NULL pointer if you
// don't care about the clut)


int GetTGASize(unsigned char *tga, long *w, long *h)
{
  TGAHEAD hdr;
  
  if (!tga) return 0;

  hdr.id_len      = *tga++;
  hdr.cmap_type   = *tga++;
  hdr.img_type    = *tga++;
  hdr.cmap_origin = tga[0] | (tga[1] << 8); tga += 2;
  hdr.cmap_len    = tga[0] | (tga[1] << 8); tga += 2;
  hdr.cdepth      = tga[0] ; tga += 1;
  hdr.xorigin     = hdr.yorigin = 0; tga += 4;  // ignore the origin field
  hdr.width       = tga[0] | (tga[1] << 8); tga += 2;
  hdr.height      = tga[0] | (tga[1] << 8); tga += 2;
  hdr.pixel_size  = *tga++;
  hdr.descrip     = *tga++;

  *w = hdr.width;
  *h = hdr.height;
	
  return hdr.cdepth;
}

int ReadTGA(unsigned char *tga, long dmaflags, void *base,
        int xOutOffset, int yOutOffset, unsigned long *clut, long tcolRGB)
{
  TGAHEAD hdr;
  unsigned int c;

  unsigned long localclut[PALETTE_SIZE];
  int fixedalpha = 0;

  int starty, endy, ystep;
  int pixcount;

  int i;
  int is16bpp, istruecolor;
  int y, x;

  
  tColor = tcolRGB;

  if (!tga) {
    return 0;
  }

  if (!clut) {
    clut = localclut;
  }
  
  hdr.id_len      = *tga++;
  hdr.cmap_type   = *tga++;
  hdr.img_type    = *tga++;
  hdr.cmap_origin =  tga[0] | (tga[1] << 8); tga += 2;
  hdr.cmap_len    =  tga[0] | (tga[1] << 8); tga += 2;
  hdr.cdepth      =  tga[0] ; tga += 1;
  hdr.xorigin     =  hdr.yorigin = 0; tga += 4;  // ignore the origin field
  hdr.width       =  tga[0] | (tga[1] << 8); tga += 2;
  hdr.height      =  tga[0] | (tga[1] << 8); tga += 2;
  hdr.pixel_size  = *tga++;
  hdr.descrip     = *tga++;

  tga += hdr.id_len; // skip identifier

  if (hdr.cmap_type != 1) {
#ifdef DEBUG
    printf("Error: TGA file not color mapped\n");
#endif
    return 0;
  }
  if (hdr.img_type != 9) {
#ifdef DEBUG
    printf("Error: TGA file not RLE encoded\n");
#endif
    return 0;
  }
  if (hdr.cdepth == 24) {
    fixedalpha = 1;        // targa file contains no alphas
  } else if (hdr.cdepth == 32) {
    fixedalpha = 0;        // alphas come from targa file
  } else {
#ifdef DEBUG
    printf("TGA Error: depth of color map entries is %d\n", hdr.cdepth);
#endif
    return 0;
  }

  /* figure out if the destination is a truecolor bitmap, and if
     so, whether it is a 16bpp or 32bpp bitmap */
  {
    int pixtype = (dmaflags >> 4) & 0xf;
    if (pixtype == 3) {  // 8bpp
      istruecolor = is16bpp = FALSE;
    } else {
      istruecolor = TRUE;
      is16bpp = (pixtype != 4);
    }
  }

  /* read the CLUT data; at most 256 entries are allowed */
  if (hdr.cmap_len > PALETTE_SIZE) {
#ifdef DEBUG
    printf("TGA Error: too many CLUT entries (%d)\n", hdr.cmap_len);
#endif
    return 0;
  }

  if (is16bpp) {
    // make CLUT entries 16bpp pixels
    for (i = 0; i < hdr.cmap_len; i++) {
      clut[i] = YCC16FromRGB(tga[2], tga[1], tga[0]);
      tga += (fixedalpha) ? 3 : 4;
    }
  } else {
    if (fixedalpha) {
      // read 24 bit CLUT entries
      for (i = 0; i < hdr.cmap_len; i++) {
	//	    clut[i] = My_YCCAFromRGBA(tga[2], tga[1], tga[0], 0);
	clut[i] = My_YCCAFromRGBA(tga[2], tga[1], tga[0], 255);
	//	    clut[i] = YCCAFromRGBA(tga[2], tga[1], tga[0], 255);
	tga += 3;
      }
    } else {
      // read 32 bit CLUT entries
      for (i = 0; i < hdr.cmap_len; i++) {
	//	    clut[i] = My_YCCAFromRGBA(tga[2], tga[1], tga[0], tga[3]);
	clut[i] = My_YCCAFromRGBA(tga[2], tga[1], tga[0], 255-tga[3]);
	//	    clut[i] = YCCAFromRGBA(tga[2], tga[1], tga[0], tga[3]);
	tga += 4;
      }
    }
  }
    
  /* check descriptor byte to see which order the image is stored in */
  if (hdr.descrip & (1<<5)) {
    starty = yOutOffset;
    endy   = hdr.height + yOutOffset;
    ystep  = 1;
  } else {
    starty = (hdr.height-1) + yOutOffset;
    endy   = yOutOffset - 1;
    ystep  = -1;
  }

  if (hdr.width > MAX_WIDTH) {
#ifdef DEBUG
    printf("TGA Error: file too wide (%d)\n", hdr.width);
#endif
    return 0;
  }
  
  if ((hdr.width & 1) && !istruecolor) {
#ifdef DEBUG
    printf("TGA Error: odd width (%d)\n", hdr.width);
#endif
    return 0;
  }

  y = starty;
  x = 0;

  if (istruecolor) {
    do {
      // get next RLE packet
      c = *tga++;
      if (c & 0x80) {
	/* one pixel repeated */
	i = (c - 0x80)+1;  /* repeat count */
	//printf("\nRLE(%d): ", i);
	//printf("%02X", *tga);
	c = clut[*tga++];
	while (i > 0) {
	  pixcount = (i > 32) ? 32 : i;
	  if (x + pixcount > hdr.width)
	    pixcount = hdr.width - x;
	  _raw_plotpixel(dmaflags, base, (pixcount<<16)|(x+xOutOffset), (1<<16)|y, c);
	  x += pixcount;
	  i -= pixcount;
	  if (x >= hdr.width) {
	    x = 0;
	    y += ystep;
	    if (y == endy) {
	      break;
	    }
	  }
	}
      } else {
	/* a bunch of unique pixels */
	i = c+1;
	//printf("\nPIX(%d): ", i);
	while (i-- > 0) {
	  //printf("%02X ", tga[0]);
	  _raw_plotpixel(dmaflags, base, (1<<16)|(x+xOutOffset), (1<<16)|y, clut[tga[0]]);
	  tga++;
	  x++;
	  if (x >= hdr.width) {
	    x = 0;
	    y += ystep;
	    if (y == endy) {
	      break;
	    }
	  }
	}
      }
    } while (y != endy);
  } else {
    // 8bpp -- have to output 2 pixels at a time
    int lastc;
      
    lastc = -1;
    do {
      // get next RLE packet
      c = *tga++;
      if (c & 0x80) {
	/* one pixel repeated */
	i = (c - 0x80)+1;  /* repeat count */
	c = *tga++;
	while (i-- > 0) {
	  if (lastc < 0) {
	    lastc = c;
	  } else {
	    _raw_plotpixel(dmaflags, base, (2<<16)|(x+xOutOffset), (1<<16)|y,
			   (lastc << 24) | (c << 16));
	    x += 2;
	    lastc = -1;
	    if (x >= hdr.width) {
	      x = 0;
	      y += ystep;
	      if (y == endy) {
		break;
	      }
	    }
	  }
	}
      } else {
	/* a bunch of unique pixels */
	i = c+1;
	while (i-- > 0) {
	  if (lastc < 0) {
	    lastc = *tga++;
	  } else {
	    c = *tga++;
	    _raw_plotpixel(dmaflags, base, (2<<16)|(x+xOutOffset), (1<<16)|y,
			   (lastc << 24) | (c<<16));
	    x+= 2;
	    lastc = -1;
	    if (x >= hdr.width) {
	      x = 0;
	      y += ystep;
	      if (y == endy) {
		break;
	      }
	    }
	  }
	}
      }
    } while (y != endy);
  }

  return 1;
}


int ReadTGAToMem32bpp(unsigned char *tga, void *base, long tcolRGB)
{
    TGAHEAD hdr;
    unsigned int c;

    //    int xOutOffset = 0;
    int yOutOffset = 0;

    unsigned long clut[PALETTE_SIZE];
    int fixedalpha = 0;

    int starty, endy, ystep;
    int pixcount;

    int i;
    int is16bpp, istruecolor;
    int y, x;

    unsigned long *wrtPtr = (unsigned long *) base;


    tColor = tcolRGB;

    if (!tga) return 0;

    hdr.id_len      = *tga++;
    hdr.cmap_type   = *tga++;
    hdr.img_type    = *tga++;
    hdr.cmap_origin = tga[0] | (tga[1] << 8); tga += 2;
    hdr.cmap_len    = tga[0] | (tga[1] << 8); tga += 2;
    hdr.cdepth      = tga[0] ; tga += 1;
    hdr.xorigin     = hdr.yorigin = 0; tga += 4;  // ignore the origin field
    hdr.width       = tga[0] | (tga[1] << 8); tga += 2;
    hdr.height      = tga[0] | (tga[1] << 8); tga += 2;
    hdr.pixel_size  = *tga++;
    hdr.descrip     = *tga++;
    tga += hdr.id_len; // skip identifier

    if (hdr.cmap_type != 1) {
#ifdef DEBUG
        printf("Error: TGA file not color mapped\n");
#endif
        return 0;
    }
    if (hdr.img_type != 9) {
#ifdef DEBUG
        printf("Error: TGA file not RLE encoded\n");
#endif
        return 0;
    }
    if (hdr.cdepth == 24) {
        fixedalpha = 1;        // targa file contains no alphas
    } else if (hdr.cdepth == 32) {
        fixedalpha = 0;        // alphas come from targa file
    } else {
#ifdef DEBUG
        printf("TGA Error: depth of color map entries is %d\n", hdr.cdepth);
#endif
        return 0;
    }

    /* fixed in this routine */
    istruecolor = TRUE;
    is16bpp     = FALSE;

    /* read the CLUT data; at most 256 entries are allowed */
    if (hdr.cmap_len > PALETTE_SIZE) {
#ifdef DEBUG
        printf("TGA Error: too many CLUT entries (%d)\n", hdr.cmap_len);
#endif
        return 0;
    }

    if (is16bpp) {
      // make CLUT entries 16bpp pixels
        for (i = 0; i < hdr.cmap_len; i++) {
	  clut[i] = YCC16FromRGB(tga[2], tga[1], tga[0]);
	  tga += (fixedalpha) ? 3 : 4;
        }
    } else {
        if (fixedalpha) {
	  // read 24 bit CLUT entries
	  for (i = 0; i < hdr.cmap_len; i++) {
	    //	    clut[i] = My_YCCAFromRGBA(tga[2], tga[1], tga[0], 0);
	    clut[i] = My_YCCAFromRGBA(tga[2], tga[1], tga[0], 255);
	    //	    clut[i] = YCCAFromRGBA(tga[2], tga[1], tga[0], 255);
	    tga += 3;
	  }
        } else {
	  // read 32 bit CLUT entries
	  for (i = 0; i < hdr.cmap_len; i++) {
//	    clut[i] = My_YCCAFromRGBA(tga[2], tga[1], tga[0], tga[3]);
	    clut[i] = My_YCCAFromRGBA(tga[2], tga[1], tga[0], 255-tga[3]);
//	    clut[i] = YCCAFromRGBA(tga[2], tga[1], tga[0], tga[3]);
	    tga += 4;
	  }
        }
    }
    
    /* check descriptor byte to see which order the image is stored in */
    if (hdr.descrip & (1<<5)) {
      starty = yOutOffset;
      endy = hdr.height + yOutOffset;
      ystep = 1;
    } else {
      //      starty = (hdr.height-1)+yOutOffset;
      //      endy = yOutOffset-1;
      //      ystep = -1;
#ifdef DEBUG
      printf("TGA Error: cannot handle backwards order\n");
#endif
      return 0;
    }

    if (hdr.width > MAX_WIDTH) {
#ifdef DEBUG
        printf("TGA Error: file too wide (%d)\n", hdr.width);
#endif
        return 0;
    }

    if ((hdr.width & 1) && !istruecolor) {
#ifdef DEBUG
        printf("TGA Error: odd width (%d)\n", hdr.width);
#endif
        return 0;
    }

    //    y = starty;
    y = 0;
    x = 0;

    if (istruecolor) {
      do {
	// get next RLE packet
	c = *tga++;
	if (c & 0x80) {
	  /* one pixel repeated */
	  i = (c - 0x80)+1;  /* repeat count */
	  c = clut[*tga++];
	  while (i > 0) {
	    pixcount = (i > 32) ? 32 : i;
	    if (x + pixcount > hdr.width)
	      pixcount = hdr.width - x;
	    //	    _raw_plotpixel(dmaflags, base, (pixcount<<16)|(x+xOutOffset), (1<<16)|y, c);
	    *wrtPtr = c;
	    wrtPtr++;
	    x += pixcount;
	    i -= pixcount;
	    if (x >= hdr.width) {
	      x = 0;
	      y += ystep;
	      if (y == endy) {
		break;
	      }
	    }
	  }
	} else {
	  /* a bunch of unique pixels */
	  i = c+1;
	  while (i-- > 0) {
	    //	    _raw_plotpixel(dmaflags, base, (1<<16)|(x+xOutOffset), (1<<16)|y, clut[tga[0]]);
	    *wrtPtr = c;
	    wrtPtr++;
	    tga++;
	    x++;
	    if (x >= hdr.width) {
	      x = 0;
	      y += ystep;
	      if (y == endy) {
		break;
	      }
	    }
	  }
	}
      } while (y != endy);
    }
    
    return 1;
}



int TGAReadHeader(unsigned char *tga, TGAHEAD *head)
{
  head->id_len      = *tga++;
  head->cmap_type   = *tga++;
  head->img_type    = *tga++;
  head->cmap_origin =  tga[0] | (tga[1] << 8); tga += 2;
  head->cmap_len    =  tga[0] | (tga[1] << 8); tga += 2;
  head->cdepth      =  tga[0] ; tga += 1;
  head->xorigin     =  head->yorigin = 0; tga += 4;  // ignore the origin field
  head->width       =  tga[0] | (tga[1] << 8); tga += 2;
  head->height      =  tga[0] | (tga[1] << 8); tga += 2;
  head->pixel_size  = *tga++;
  head->descrip     = *tga++;

  if (head->cmap_type != 1) {
    printf("TGA ERROR: cmap_type=%d\n", head->cmap_type);
    return 1;
  }
  if ((head->img_type & 0x7) != 1) {
    printf("TGA ERROR: img_type=%d\n", head->img_type);
    return 1;
  }
  if (head->pixel_size != 8) {
    printf("TGA ERROR: pixel_size=%d\n", head->pixel_size);
    return 1;
  }
  if ((head->cdepth != 24) && (head->cdepth != 32)) {
    printf("TGA ERROR: cdepth=%d\n", head->cdepth);
    return 1;
  }

  return 0;
}

void TGAReadClut(TGA *tga, int use16bpp)
{
  int i;
  unsigned char *ptr;

  if (!tga) {
    return;
  }

  if (!tga->clut) {
    return;
  }

  if (tga->head.cdepth == 24) {
    tga->bytesPerColor = 3;
  } else if (tga->head.cdepth == 32) {
    tga->bytesPerColor = 4;
  }

  ptr = tga->ptr + 18 + tga->head.id_len;

  // Loop through the colors, converting the RGB to YCrCb
  if (use16bpp) {
    // make CLUT entries 16bpp pixels
    for (i = 0; i < tga->head.cmap_len; i++) {
      tga->clut[i + tga->head.cmap_origin] = YCC16FromRGB(ptr[2], ptr[1], ptr[0]);
      ptr += tga->bytesPerColor;
    }

  } else {
    if (tga->bytesPerColor == 3) {
      // read 24 bit CLUT entries
      for (i = 0; i < tga->head.cmap_len; i++) {
	tga->clut[i + tga->head.cmap_origin] = My_YCCAFromRGBA(ptr[2], ptr[1], ptr[0], 255);
	ptr += 3;
      }
    } else {
      // read 32 bit CLUT entries
      for (i = 0; i < tga->head.cmap_len; i++) {
	tga->clut[i + tga->head.cmap_origin] = My_YCCAFromRGBA(ptr[2], ptr[1], ptr[0], 255-ptr[3]);
	ptr += 4;
      }
    }
  }
}

TGA* TGAInit(void *tgaPtr)
{
  TGA *tga;

  tga = (TGA*) malloc(sizeof(TGA));
  if (!tga) {
    printf("ERROR - Could not allocate new TGA structure\n");
    return NULL;
  }

  tga->ptr = tgaPtr;
  if (TGAReadHeader(tgaPtr, &tga->head)) {
    free(tga);
    return NULL;
  }

  tga->clut = (unsigned long*) malloc(sizeof(unsigned long) * PALETTE_SIZE);
  if (!tga->clut) {
    printf("ERROR - Could not allocate memory for clut\n");
    free(tga);
    return NULL;
  }

  TGAReadClut(tga, FALSE);

  tga->curY = 0;
  tga->curPtr = tga->ptr + 18 + tga->head.id_len + (tga->bytesPerColor * tga->head.cmap_len);
  tga->curOffset = 0;

  return tga;
}

int TGAReadLine(TGA *tga, long dmaflags, void *base, int xOffset, int yOffset)
{
  int i, x = xOffset, y;
  unsigned char c;
  unsigned long color;
  int pixelsToSkip = tga->curOffset, pixelsToSkipPrev;
  int pixelsToWrite;
  int pixelsWrittenThisPacket = 0;
  int pixelsWrittenThisLine   = 0;
  unsigned char *ptr = tga->curPtr;
  unsigned long *localPtr;
  int offset;
  int fullPacket;
  int debug = 0;

  localScratch = _MemLocalScratch(&localSize);
  localPtr = (unsigned long *) localScratch;

  // Get destination Y coordinate
  if (tga->head.descrip & (1<<5)) {
    y = tga->curY;
  } else {
    y = tga->head.height - 1 - tga->curY;
  }
  y += yOffset;

  if (debug) {
    printf("%s: y=%d  (%02X %08lX)\n", __func__, y, tga->head.descrip, dmaflags);
  }

  // Loop until the line is done
  while (pixelsWrittenThisLine < tga->head.width) {
    //
    // Figure out what type of packet this is and how many pixels to write.
    //
    c = ptr[0];
    if (debug) printf("%s: New packet=%02X\n", __func__, c);

    if (c & 0x80) {
      //
      // One pixel repeated
      //

      if (debug) printf("%s: Repeat count=%d\n", __func__, (c - 0x80) + 1);

      // Get repeat count
      pixelsToWrite = (c - 0x80) + 1 - pixelsToSkip;

      if (debug) printf("%s: Pixels to write=%d\n", __func__, pixelsToWrite);

      //if (debug) printf("\nRLE(%d): ", pixelsToWrite);
      //if (debug) printf("%02X", ptr[1]);
    } else {
      //
      // A bunch of unique pixels
      //

      if (debug) printf("%s: Num unique pixels=%d\n", __func__, c + 1);

      // Get number of pixels
      pixelsToWrite = c + 1 - pixelsToSkip;

      if (debug) printf("%s: Pixels to write=%d\n", __func__, pixelsToWrite);

      //if (debug) printf("\nPIX(%d): ", pixelsToWrite);
    }

    pixelsToSkipPrev = pixelsToSkip;

    fullPacket = TRUE;
    if ((pixelsWrittenThisPacket + pixelsToWrite) > 64) {
      pixelsToWrite = 64 - pixelsWrittenThisPacket;
      pixelsToSkip += pixelsToWrite;
      fullPacket = FALSE;
      if (debug) printf("%s: Saturate to packet size - Pixels to write=%d\n", __func__, pixelsToWrite);
    }

    if ((pixelsWrittenThisLine + pixelsToWrite) > tga->head.width) {
      pixelsToWrite = tga->head.width - pixelsWrittenThisLine;
      pixelsToSkip += pixelsToWrite;
      fullPacket = FALSE;
      if (debug) printf("%s: Saturate to line width - Pixels to write=%d\n", __func__, pixelsToWrite);
    }

    pixelsWrittenThisLine   += pixelsToWrite;
    pixelsWrittenThisPacket += pixelsToWrite;

    if (debug) printf("%s: Pixels written this line=%d\n", __func__, pixelsWrittenThisLine);
    if (debug) printf("%s: Pixels written this packet=%d\n", __func__, pixelsWrittenThisPacket);

    //
    // Write the pixels to the internal buffer.
    //
    if (c & 0x80) {
      //
      // One pixel repeated
      //

      // Get color
      color = tga->clut[ptr[1]];

      if (debug) printf("%s: Repeat color=%08X (%02X)\n", __func__, (int) color, ptr[1]);

      // Write the pixels to our internal buffer
      for (i = 0; i < pixelsToWrite; i++, localPtr++) {
	*localPtr = color;
      }

      // See if we're done with this repeated pixel
      if (fullPacket) {
	pixelsToSkip = 0;
	ptr += 2;  // skip to the next RLE packet
	if (debug) printf("%s: Packet done\n", __func__);
      }

    } else {
      //
      // A bunch of unique pixels
      //

      for (i = 0, offset = 1 + pixelsToSkipPrev; i < pixelsToWrite; i++, localPtr++, offset++) {
	*localPtr = tga->clut[ptr[offset]];
      }

      if (fullPacket) {
	pixelsToSkip = 0;
	ptr += c + 2;  // skip to the next RLE packet
	if (debug) printf("%s: Packet done\n", __func__);
      }
    }

    if (debug) printf("%s: Pixels to skip=%d\n", __func__, pixelsToSkip);
    
    // If we filled the buffer then blast out the pixels.
    if ((pixelsWrittenThisPacket == 64) || (pixelsWrittenThisLine == tga->head.width)) {
      _DMABiLinear(dmaflags, base, (pixelsWrittenThisPacket<<16) | x, (1<<16) | y, localScratch);
      x += pixelsWrittenThisPacket;
      localPtr = (unsigned long *) localScratch;
      pixelsWrittenThisPacket = 0;
      if (debug) printf("%s: Buffer full or line done, writing pixels\n", __func__);
    }
  }

  tga->curPtr    = ptr;
  tga->curOffset = pixelsToSkip;
  tga->curY++;

  if (debug) printf("%s: Done.  ptr=%08X  offset=%d  y=%d\n", __func__, (unsigned int) tga->curPtr,
		    tga->curOffset, tga->curY);

  if (tga->curY == tga->head.height) {
    // Done with all the lines
    return 1;
  }

  return 0;
}

void TGADelete(TGA *tga)
{
  if (tga) {

    if (tga->clut) {
      free(tga->clut);
    }

    free(tga);
  }
}
